'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.typeIsArray = undefined;
exports.caseInsensitiveRegex = caseInsensitiveRegex;
exports.isNullOrEmpty = isNullOrEmpty;
exports.isNullOrWhitespace = isNullOrWhitespace;
exports.logAndSetResponse = logAndSetResponse;
exports.getAllChannelsInPriorityOrder = getAllChannelsInPriorityOrder;
exports.getKeystore = getKeystore;
exports.statusCodePatternMatch = statusCodePatternMatch;
exports.uniqArray = uniqArray;
exports.serverTimezone = serverTimezone;
exports.selectAuditFields = selectAuditFields;

var _momentTimezone = require('moment-timezone');

var _momentTimezone2 = _interopRequireDefault(_momentTimezone);

var _winston = require('winston');

var _winston2 = _interopRequireDefault(_winston);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _channels = require('./model/channels');

var _keystore = require('./model/keystore');

var _config = require('./config');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

_config.config.caching = _config.config.get('caching');
_config.config.api = _config.config.get('api');

/**
 * Will take in a string and return a safe regex that will match case insensitive
 *
 * @export
 * @param {string} value that needs to be matched
 * @returns {RegExp} regex that will match case insensitive
 */
function caseInsensitiveRegex(value) {
  return new RegExp(`^${_lodash2.default.escapeRegExp(value)}$`, 'i');
}

function isNullOrEmpty(arr) {
  if (arr == null) {
    return true;
  }

  return arr.length === 0;
}

function isNullOrWhitespace(value) {
  return (/^\s*$/.test(value || '')
  );
}

// function to log errors and return response
function logAndSetResponse(ctx, status, msg, logLevel) {
  _winston2.default[logLevel](msg);
  ctx.body = msg;
  ctx.status = status;

  return status;
}

const cacheValueStore = {};

const { refreshMillis } = _config.config.caching;

function getCachedValues(store, callback) {
  const lastCheck = cacheValueStore[`${store}`] != null ? cacheValueStore[`${store}`].lastCheck : undefined;

  if (!_config.config.caching.enabled || lastCheck == null || new Date() - lastCheck > refreshMillis) {
    const handler = (err, results) => {
      if (err) {
        return callback(err);
      }

      if (_config.config.caching.enabled) {
        if (!lastCheck) {
          cacheValueStore[`${store}`] = {};
        }
        cacheValueStore[`${store}`].value = results;
        cacheValueStore[`${store}`].lastCheck = new Date();
      }

      return callback(null, results);
    };

    // TODO make this more generic (had issues passing Channel.find as a param [higher order function])
    if (store === 'channels') {
      return _channels.ChannelModel.find({}).sort({ priority: 1 }).exec((err, channels) => {
        if (err) {
          return handler(err);
        }
        const noPriorityChannels = [];
        const sortedChannels = [];
        channels.forEach(channel => {
          if (channel.priority == null) {
            return noPriorityChannels.push(channel);
          } else {
            return sortedChannels.push(channel);
          }
        });
        return handler(null, sortedChannels.concat(noPriorityChannels));
      });
    } else if (store === 'keystore') {
      return _keystore.KeystoreModel.findOne({}, handler);
    } else {
      return callback(new Error(`Internal error: Invalid store ${store}`));
    }
  } else {
    return callback(null, cacheValueStore[`${store}`].value);
  }
}

function getAllChannelsInPriorityOrder(callback) {
  return getCachedValues('channels', callback);
}

function getKeystore(callback) {
  return getCachedValues('keystore', callback);
}

// function to check if string match status code pattern
function statusCodePatternMatch(string, callback) {
  return (/\dxx/.test(string)
  );
}

// returns an array with no duplicates
function uniqArray(arr) {
  const dict = arr.reduce((p, c) => {
    p[c] = c;
    return p;
  }, {});

  const result = [];
  for (const k in dict) {
    const v = dict[k];
    result.push(v);
  }
  return result;
}

// thanks to https://coffeescript-cookbook.github.io/chapters/arrays/check-type-is-array
const typeIsArray = exports.typeIsArray = Array.isArray || (value => ({}).toString.call(value) === '[object Array]');

// get the server timezone
function serverTimezone() {
  return _momentTimezone2.default.tz.guess();
}

/**
 * Return an object containing the relevant fields for audit logging from the authenticated user.
 *
 * @param {Object} authenticated The authenticated user.
 * @return {Object} The object containing selected audit fields.
 */
function selectAuditFields(authenticated) {
  return {
    id: authenticated._id,
    name: `${authenticated.firstname} ${authenticated.surname}`
  };
}
//# sourceMappingURL=utils.js.map